/*
    Copyright (C) 2001-2003 KSVG Team
    This file is part of the KDE project

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include <kdebug.h>
#include <qimage.h>

#include "SVGMaskElement.h"

#include "SVGRectImpl.h"
#include "SVGDocumentImpl.h"
#include "SVGSVGElementImpl.h"
#include "SVGMaskElementImpl.h"
#include "SVGAnimatedLengthImpl.h"
#include "SVGAnimatedEnumerationImpl.h"
#include "SVGUnitConverter.h"
#include "SVGShapeImpl.h"
#include "SVGMatrixImpl.h"
#include "KSVGCanvas.h"
#include "CanvasItems.h"
#include "CanvasFactory.h"

using namespace KSVG;

#include "SVGMaskElementImpl.lut.h"
#include "ksvg_scriptinterpreter.h"
#include "ksvg_ecma.h"

SVGMaskElementImpl::SVGMaskElementImpl(DOM::ElementImpl *impl) : SVGElementImpl(impl), SVGTestsImpl(), SVGLangSpaceImpl(), SVGExternalResourcesRequiredImpl(), SVGStylableImpl(this), SVGBBoxTarget()
{
	KSVG_EMPTY_FLAGS

	m_x = new SVGAnimatedLengthImpl();
	m_x->ref();

	m_y = new SVGAnimatedLengthImpl();
	m_y->ref();

	m_width = new SVGAnimatedLengthImpl();
	m_width->ref();

	m_height = new SVGAnimatedLengthImpl();
	m_height->ref();

	m_maskUnits = new SVGAnimatedEnumerationImpl();
	m_maskUnits->ref();

	m_maskContentUnits = new SVGAnimatedEnumerationImpl();
	m_maskContentUnits->ref();

	m_converter = new SVGUnitConverter();
	m_converter->add(m_x);
	m_converter->add(m_y);
	m_converter->add(m_width);
	m_converter->add(m_height);

	m_canvas = 0;
}

SVGMaskElementImpl::~SVGMaskElementImpl()
{
	if(m_x)
		m_x->deref();
	if(m_y)
		m_y->deref();
	if(m_width)
		m_width->deref();
	if(m_height)
		m_height->deref();
	if(m_maskUnits)
		m_maskContentUnits->deref();
	if(m_maskUnits)
		m_maskContentUnits->deref();
	delete m_converter;
	if(m_canvas)
		delete m_canvas;
}

SVGAnimatedEnumerationImpl *SVGMaskElementImpl::maskUnits() const
{
	return m_maskUnits;
}

SVGAnimatedEnumerationImpl *SVGMaskElementImpl::maskContentUnits() const
{
	return m_maskContentUnits;
}

SVGAnimatedLengthImpl *SVGMaskElementImpl::x() const
{
	return m_x;
}

SVGAnimatedLengthImpl *SVGMaskElementImpl::y() const
{
	return m_y;
}

SVGAnimatedLengthImpl *SVGMaskElementImpl::width() const
{
	return m_width;
}

SVGAnimatedLengthImpl *SVGMaskElementImpl::height() const
{
	return m_height;
}

/*
@begin SVGMaskElementImpl::s_hashTable 7
 maskUnits			SVGMaskElementImpl::MaskUnits			DontDelete|ReadOnly
 maskContentUnits	SVGMaskElementImpl::MaskContentUnits	DontDelete|ReadOnly
 x					SVGMaskElementImpl::X					DontDelete|ReadOnly
 y					SVGMaskElementImpl::Y					DontDelete|ReadOnly
 width				SVGMaskElementImpl::Width				DontDelete|ReadOnly
 height				SVGMaskElementImpl::Height				DontDelete|ReadOnly
@end
*/
Value SVGMaskElementImpl::getValueProperty(ExecState *exec, int token) const
{
	KSVG_CHECK_ATTRIBUTE
		
	switch(token)
	{
		case MaskUnits:
			if(!attributeMode)
				return m_maskUnits->cache(exec);
			else
				return Number(m_maskUnits->baseVal());
		case MaskContentUnits:
			if(!attributeMode)
				return m_maskContentUnits->cache(exec);
			else
				return Number(m_maskContentUnits->baseVal());
		case X:
			if(!attributeMode)
				return m_x->cache(exec);
			else
				return Number(m_x->baseVal()->value());
		case Y:
			if(!attributeMode)
				return m_y->cache(exec);
			else
				return Number(m_y->baseVal()->value());
		case Width:
			if(!attributeMode)
				return m_width->cache(exec);
			else
				return Number(m_width->baseVal()->value());

		case Height:
			if(!attributeMode)
				return m_height->cache(exec);
			else
				return Number(m_height->baseVal()->value());
		default:
			kdWarning() << "Unhandled token in " << k_funcinfo << " : " << token << endl;
			return Undefined();
	}
}

void SVGMaskElementImpl::putValueProperty(ExecState *exec, int token, const Value &value, int attr)
{
    // This class has just ReadOnly properties, only with the Internal flag set
    // it's allowed to modify those.
    if(!(attr & KJS::Internal))
        return;

    switch(token)
    {
        case MaskUnits:
            if(value.toString(exec).qstring() == "objectBoundingBox")
                m_maskUnits->setBaseVal(SVGMaskElement::SVG_UNIT_TYPE_OBJECTBOUNDINGBOX);
            else
                m_maskUnits->setBaseVal(SVGMaskElement::SVG_UNIT_TYPE_USERSPACEONUSE);
            break;
        case MaskContentUnits:
            if(value.toString(exec).qstring() == "objectBoundingBox")
                m_maskContentUnits->setBaseVal(SVGMaskElement::SVG_UNIT_TYPE_OBJECTBOUNDINGBOX);
            else
                m_maskContentUnits->setBaseVal(SVGMaskElement::SVG_UNIT_TYPE_USERSPACEONUSE);
            break;
		case X:
			converter()->modify(x(), value.toString(exec).qstring());
			break;
		case Y:
			converter()->modify(y(), value.toString(exec).qstring());
			break;
		case Width:
			converter()->modify(width(), value.toString(exec).qstring());
			break;
		case Height:
			converter()->modify(height(), value.toString(exec).qstring());
			break;
        default:
            kdWarning() << "Unhandled token in " << k_funcinfo << " : " << token << endl;
    }
}

SVGRectImpl *SVGMaskElementImpl::getBBox()
{
	SVGRectImpl *ret = SVGSVGElementImpl::createSVGRect();
	ret->setX(m_x->baseVal()->value());
	ret->setY(m_y->baseVal()->value());
	ret->setWidth(m_width->baseVal()->value());
	ret->setHeight(m_height->baseVal()->value());
	return ret;
}

void SVGMaskElementImpl::setAttributes()
{
	SVGElementImpl::setAttributes();

    // Spec: if attribute not specified, use objectBoundingBox
    if(KSVG_TOKEN_NOT_PARSED(MaskUnits))
        KSVG_SET_ALT_ATTRIBUTE(MaskUnits, "objectBoundingBox")

	// Spec: if attribute not specified, use userSpaceOnUse
    if(KSVG_TOKEN_NOT_PARSED(MaskContentUnits))
        KSVG_SET_ALT_ATTRIBUTE(MaskContentUnits, "userSpaceOnUse")

	// Spec: if attribute not specified, use "-10%"
    if(KSVG_TOKEN_NOT_PARSED(X))
        KSVG_SET_ALT_ATTRIBUTE(X, "-10%");

	// Spec: if attribute not specified, use "-10%"
    if(KSVG_TOKEN_NOT_PARSED(Y))
        KSVG_SET_ALT_ATTRIBUTE(Y, "-10%");

	// Spec: if attribute not specified, use "120%"
    if(KSVG_TOKEN_NOT_PARSED(Width))
        KSVG_SET_ALT_ATTRIBUTE(Width, "120%");

	// Spec: if attribute not specified, use "120%"
    if(KSVG_TOKEN_NOT_PARSED(Height))
        KSVG_SET_ALT_ATTRIBUTE(Height, "120%");
}

QImage SVGMaskElementImpl::createMask(SVGShapeImpl *referencingElement)
{
	converter()->finalize(referencingElement, ownerSVGElement(), maskUnits()->baseVal());

	int imageWidth = static_cast<int>(width()->baseVal()->value() + 0.5);
	int imageHeight = static_cast<int>(height()->baseVal()->value() + 0.5);

	QImage image(imageWidth, imageHeight, 32);
	image.setAlphaBuffer(true);

	if(m_canvas == 0)
		m_canvas = CanvasFactory::request()->loadCanvas(image.width(), image.height());

	m_canvas->setup(image.bits(), image.width(), image.height());
	m_canvas->setBackgroundColor(qRgba(0, 0, 0, 0));

	SVGMatrixImpl *baseMatrix = SVGSVGElementImpl::createSVGMatrix();

	// Set the scale to map the tile onto the integral sized image
	double xScale = static_cast<double>(imageWidth) / width()->baseVal()->value();
	double yScale = static_cast<double>(imageHeight) / height()->baseVal()->value();

	//kdDebug() << "SCALE " << xScale << " " << yScale << endl;
	//kdDebug() << "INT SIZE " << imageWidth << " " << imageHeight << endl;

	baseMatrix->scaleNonUniform(xScale, yScale);

	kdDebug() << "x " << x()->baseVal()->value() << " y " << y()->baseVal()->value() << endl;

	baseMatrix->translate(-x()->baseVal()->value(), -y()->baseVal()->value());

	if(maskContentUnits()->baseVal() == SVGMaskElement::SVG_UNIT_TYPE_OBJECTBOUNDINGBOX)
	{
		// Get local coordinate bounding box
		SVGRectImpl *rect = referencingElement->getBBox();
		
		baseMatrix->translate(rect->x(), rect->y());
		baseMatrix->scaleNonUniform(rect->width(), rect->height());
		rect->deref();
	}

	for(DOM::Node node = firstChild(); !node.isNull(); node = node.nextSibling())
	{
		SVGElementImpl *element = ownerDoc()->getElementFromHandle(node.handle());
		SVGShapeImpl *shape = dynamic_cast<SVGShapeImpl *>(element);
		SVGTestsImpl *tests = dynamic_cast<SVGTestsImpl *>(element);
		SVGStylableImpl *style = dynamic_cast<SVGStylableImpl *>(element);

		bool ok = tests ? tests->ok() : true;
		if(element && shape && style && ok && style->getVisible() && style->getDisplay())
		{
			SVGLocatableImpl *locatable = dynamic_cast<SVGLocatableImpl *>(element);
			if(locatable)
				locatable->updateCachedScreenCTM(baseMatrix);

			element->create(m_canvas);
			if(shape->item())
			{
				shape->item()->setReferenced(true);
				m_canvas->invalidate(shape->item(), true);
			}
		}
	}

	baseMatrix->deref();

	m_canvas->setPan(QPoint(0,0));
	m_canvas->update(float(1));

	//image.save("mask.png", "PNG");

	return image;
}

// vim:ts=4:noet
